import java.applet.Applet;
import java.awt.*;
import java.awt.event.*;
import java.awt.font.*;
import java.io.*;
import java.net.*;


public class MPatience extends Applet implements Runnable, MouseListener, MouseMotionListener {
	static final double FRAME = 10;
	static final double RATE = FRAME/1000;

	static final int SCREEN_W = 923;
	static final int SCREEN_H = 550;
	
	static final int CARD_CX = (int)(SCREEN_W / 2 - 49);
	static final int CARD_CY = (int)(SCREEN_H / 2 - 70);
	static final int CARD_AX = (int)(SCREEN_W / 2 - 59);
	static final int CARD_AY = (int)(SCREEN_H / 2 - 80);
	
	static final int XOFF1 = (int)((SCREEN_W - 559) / 14);
	static final int XOFF2 = (int)((SCREEN_W - 118) / 7);
	
	static final int T_DECK = 0;
	static final int T_WORKING = 1;
	static final int T_DONE = 2;
	static final int T_MOVING = 3;
	
	static final int S_WORKING = 1;
	static final int S_DONE = 9;
	static final int S_DECK = 0;
	static final int S_MOVING = 10;
	
	static final Font F_TITLE = new Font("SansSerif", Font.BOLD, 14);
	static final Font F_NORMAL = new Font("SansSerif", Font.PLAIN, 14);
	static final Font F_TINY = new Font("SansSerif", Font.PLAIN, 9);
	
	static final String[] T_LINES = {"The objective is to remove all eight suits from play. The game follows these simple rules:", " - Any sequential run of cards of the same suit may be moved as a single card;", " - Cards can be moved onto any card one higher in sequence, regardless of suit;", " - Any card can be moved into an empty column;", " - Completed suits are removed from play.", "", "Click on the deck at the bottom right to deal", "another row of cards.", "You can backtrack as far as you like by clicking", "on the relevant button.", "You can abandon a game and start a new one", "using the 'Redeal' button.", "", "This game is very difficult to get out.", "It truly requires Patience!"};
	static final String[] T_ORDER = {"Essence", "Deity", "Six", "Five", "Four", "Three", "Two", "One"};
	static final String[] T_SUITS = {"Earth", "Water", "Air", "Fire", "Lightning", "Force", "Light", "Psi"};
	
	
	Thread theThread;
	int mode = 9999;
	double time = 0;
	double timecheck = 0;
	boolean drawing = false;

	MediaTracker tracker;

	Image backBuffer;
	Graphics2D backGC;
	Image[][] cards = new Image[8][8];
	Image back;
	Image stack;
	Image ctls;
	Image info;
	
	Stack[] stacks = new Stack[S_MOVING + 1];
	
	State history;

	boolean test = false;
	
	
	int m_i = 0;
	int m_j = 0;
	int last_stack = -1;
	
	
	int f_px = 0;
	int f_py = 0;
	int f_sx = 0;
	int f_sy = 0;
	int f_dx = 0;
	int f_dy = 0;
	
	

	public void init() {
		loadImages();
		
		backBuffer = createImage(SCREEN_W, SCREEN_H);
		backGC = (Graphics2D)backBuffer.getGraphics();
		
		init_stacks();
		
		addMouseListener(this);
		addMouseMotionListener(this);
	}
	
	
	
	public void init_stacks() {
		history = null;
		
		for (int i=0; i<8; i++) stacks[i + S_WORKING] = new Stack(i, T_WORKING);
		
		stacks[S_DONE] = new Stack(0, T_DONE);
		stacks[S_DECK] = new Stack(0, T_DECK);
		stacks[S_MOVING] = new Stack(0, T_MOVING);
		
		Card[] s = new Card[64];
		
		for (int rank=0; rank<8; rank++)
			for (int suit=0; suit<8; suit++)
				s[suit + rank*8] = new Card(suit, rank, false);
		
		if (!test) {
			for (int i=0; i<64; i++) {
				int j = (int)Math.floor(Math.random() * 64);
				
				Card t = s[i];
				s[i] = s[j];
				s[j] = t;
			}
		}
		
		for (int i=0; i<64; i++)
			stacks[S_DECK].add(s[i]);
	}




	void loadImages() {
		tracker = new MediaTracker(this);

		try {
			for (int suit=1; suit<9; suit++)
				for (int rank=1; rank<9; rank++) {
					String tag = "" + rank + suit;
					int itag = rank*10 + suit;
					
					showStatus("Loading image " + tag + "...");
					cards[suit-1][rank-1] = getImage(new URL(getDocumentBase(), tag + ".png"));
					tracker.addImage(cards[suit-1][rank-1], itag);
					tracker.waitForID(itag);

					if (tracker.isErrorID(itag)) {
						showStatus("Error loading image " + tag);
						mode = 9999;
					}
				}
			
			showStatus("Loading image ii...");
			info = getImage(new URL(getDocumentBase(), "ii.png"));
			tracker.addImage(back, 90);
			tracker.waitForID(90);

			if (tracker.isErrorID(90)) {
				showStatus("Error loading image ii");
				mode = 9999;
			}
			
			showStatus("Loading image xx...");
			back = getImage(new URL(getDocumentBase(), "xx.png"));
			tracker.addImage(back, 100);
			tracker.waitForID(100);

			if (tracker.isErrorID(100)) {
				showStatus("Error loading image xx");
				mode = 9999;
			}
			
			showStatus("Loading image yy...");
			stack = getImage(new URL(getDocumentBase(), "yy.png"));
			tracker.addImage(stack, 110);
			tracker.waitForID(110);

			if (tracker.isErrorID(110)) {
				showStatus("Error loading image yy");
				mode = 9999;
			}
			
			showStatus("Loading image zz...");
			ctls = getImage(new URL(getDocumentBase(), "zz.png"));
			tracker.addImage(stack, 120);
			tracker.waitForID(120);

			if (tracker.isErrorID(120)) {
				showStatus("Error loading image zz");
				mode = 9999;
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	
	
	

	public void start() {
		if (theThread == null) {
			theThread = new Thread(this);
		}

		theThread.start();
	}

	public void run() {
		mode = 0;

		long startTime = System.currentTimeMillis();

		while (theThread == Thread.currentThread()) {
			// Timing...

			try {
				startTime += FRAME;
				time += RATE;

				for (int i=0; i<=S_MOVING; i++) stacks[i].move();
				
//System.out.println("Mode " + mode);
				
				switch (mode) {
					case 0:
						if (time > timecheck) {
							if ((m_i == 0) && (m_j == 0)) init_stacks();
							
							stacks[S_DECK].deal((m_j > 2) ? true : test, stacks[S_WORKING + m_i]);
							m_i++;
							
							if (m_i > 7) {
								m_i = 0;
								m_j++;
								
								if (m_j > 3) {
									for (int i=0; i<8; i++) stacks[S_WORKING + i].check();
									if (mode != 2) mode = 1;
									test = false;
									State nh = new State(stacks);
									nh.prev = history;
									history = nh;
//histdump("after deal");
								}
							}
							
							if (mode != 2) timecheck = time + 0.1;
						}
					break;
					case 1:
					break;
					case 2:
					break;
					case 3:
					break;
				}
				
				Thread.sleep(Math.max(0, startTime-System.currentTimeMillis()));
			} catch (InterruptedException e) {
				break;
			}

			repaint();
		}

		mode = 9999;
	}

	public void stop() {
		theThread = null;
	}

    public void destroy() {
		removeMouseListener(this);
		removeMouseMotionListener(this);
    }

	public void mouseClicked(MouseEvent e) {
		int x = e.getX();
		int y = e.getY();
//histdump("preclick");
		if ((y >= SCREEN_H - 150) && (y <= SCREEN_H - 10) && (x >= SCREEN_W - 83 - (int)Math.floor(stacks[S_DECK].num/8) * XOFF1) && (x <= SCREEN_W - 10) && (mode == 1)) {								// DECK HIT
			mode = 0;
			timecheck = 0;
		} else if ((y >= SCREEN_H - 120) && (y <= SCREEN_H - 40)) {
			if ((x >= SCREEN_W / 2 - 142) && (x <= SCREEN_W / 2 - 55)) {				// INFO HIT
				if (mode == 1) mode = 3;
				else if (mode == 3) mode = 1;
			} else if ((x >= SCREEN_W / 2 - 44) && (x <= SCREEN_W / 2 + 44)) {			// UNDO HIT
				if ((mode == 1) && (history != null) && (history.prev != null)) {
					history = history.prev;
					for (int i=0; i<=S_MOVING; i++) stacks[i] = history.stacks[i].copy();
				}
			} else if ((x >= SCREEN_W / 2 + 55) && (x <= SCREEN_W / 2 + 142)) {			// RESTART HIT
				if (mode < 3) {
					for (int i=S_WORKING; i<=S_MOVING; i++) stacks[i].transfer(0, stacks[S_DECK]);
					
					test = e.isAltDown();
					
					m_i = 0;
					m_j = 0;
					mode = 0;
					timecheck = time + 0.5;
				}
			}
		}
//histdump("postclick");
	}

	public void mousePressed(MouseEvent e) {
		int x = e.getX();
		int y = e.getY();
//histdump("premove");
		if (mode == 1) {
			int i = x - 10;
			
			if (i < 0) return;
			if (i % XOFF2 > 98) return;
			
			i = (int)(i/XOFF2);
			
			if (stacks[S_WORKING + i].num < 1) return;
			
			int n = y - 10;
			
			if (n < 0) return;
			if (n > stacks[S_WORKING + i].num * stacks[S_WORKING + i].yoff() + 105) return;
			
			n = (int)(n/stacks[S_WORKING + i].yoff());
			if (n >= stacks[S_WORKING + i].num) n = stacks[S_WORKING + i].num - 1;
			
			boolean ok = true;
			Card c = stacks[S_WORKING + i].bottom;
			
			for (int l=0; l<n; l++) c = c.next;
			
			if (!c.visible) return;
			
			c = c.next;
			
			while (c != null) {
				if ((c.prev != null) && ((c.prev.suit != c.suit) || (c.prev.rank != c.rank + 1))) ok = false;
				c = c.next;
			}
			
			if (ok) {
				last_stack = i;
				
				stacks[S_MOVING].x = x - 49;
				stacks[S_MOVING].y = y - 10;
				
				stacks[S_WORKING + i].transfer(n, stacks[S_MOVING]);
			}
		}
	}
	
	public void mouseReleased(MouseEvent e) {
		int x = e.getX();
		int y = e.getY();

		if ((mode == 1) && (last_stack > -1)) {
			boolean ok = true;
			
			int i = x - 10;
			
			if (i < 0) ok = false;
			if (i % XOFF2 > 98) ok = false;
			
			i = (int)(i/XOFF2);
			
			if (i > 7) ok = false;
			
			int n = (int)((y - 10)/stacks[S_WORKING + i].yoff());
			
			if (n < stacks[S_WORKING + i].num - 2) ok = false;
			if (n > stacks[S_WORKING + i].num + 2) ok = false;
			if (i == last_stack) ok = false;
			
			if ((stacks[S_WORKING + i].top != null) && (stacks[S_WORKING + i].top.rank != stacks[S_MOVING].bottom.rank + 1)) ok = false;
			
			
			if (ok) {
				if (stacks[S_WORKING + last_stack].top != null) stacks[S_WORKING + last_stack].top.visible = true;
				stacks[S_MOVING].transfer(0, stacks[S_WORKING + i]);
				stacks[S_WORKING + i].check();
				
				State nh = new State(stacks);
				nh.prev = history;
				history = nh;
			} else {
				stacks[S_MOVING].transfer(0, stacks[S_WORKING + last_stack]);
			}
			
			last_stack = -1;
		}
//histdump("postmove");
	}

	public void mouseEntered(MouseEvent e) {}
	
	public void mouseExited(MouseEvent e) {
		mouseReleased(e);
	}
	
	public void mouseDragged(MouseEvent e) {
		int x = e.getX();
		int y = e.getY();
		
		if (mode == 1) {
			stacks[S_MOVING].x = x - 49;
			stacks[S_MOVING].y = y - 10;
		}
	}
	
	public void mouseMoved(MouseEvent e) {}

	public void update(Graphics g) {
		paint(g);
	}

	public void paint(Graphics g) {
		if (!drawing) {
			drawing = true;
			

			if (mode < 9999) {
				backGC.setColor(new Color(0, 153, 0));
				backGC.fillRect(0, 0, SCREEN_W, SCREEN_H);
				backGC.drawImage(ctls, SCREEN_W/2 - 152, SCREEN_H - 130, this);
				
				for (int i=0; i<=S_MOVING; i++) stacks[i].draw(this, false);
				for (int i=0; i<=S_MOVING; i++) stacks[i].draw(this, true);
				
				if (mode == 3) {
					int ox = SCREEN_W/2 - 301;
					int oy = SCREEN_H - 530;
					
					backGC.drawImage(info, ox, oy, this);

					backGC.setColor(Color.yellow);
					
					backGC.setFont(F_TITLE);

					backGC.drawString("A Maraneldan Game of Solitaire", ox + 345, oy + 49);
					backGC.drawString("Card Order", ox + 385, oy + 174);
					backGC.drawString("Suits", ox + 507, oy + 174);
					
					backGC.setFont(F_NORMAL);
					
					for (int i=0; i<15; i++) backGC.drawString(T_LINES[i], ox + 20, oy + 90 + 19 * i);
					for (int i=0; i<8; i++) backGC.drawString(T_ORDER[i], ox + 385, oy + 199 + 25 * i);
					for (int i=0; i<8; i++) backGC.drawString(T_SUITS[i], ox + 507, oy + 199 + 25 * i);
					
					backGC.setFont(F_TINY);
					
					backGC.drawString("All icons and images are (c)2009 Maranelda Organisation.", ox + 20, oy + 380);
				}
			} else {
				showStatus("-X-X-X-  APPLET ABORTED  -X-X-X-");
			}

			g.drawImage(backBuffer, 0, 0, this);

			drawing = false;
		}
	}
	
	
	
	void histdump(String s) {
		State c = history;
		
		System.out.println("History stack (" + s + ")...");
		
		while (c != null) {
			c.desc();
			c = c.prev;
		}
	}

	
	
	class Card {
		int dx = 0;
		int dy = 0;
		int suit;
		int rank;
		boolean visible;
		Card prev;
		Card next;
		
		Card(int suit, int rank, boolean visible) {
			this.suit = suit;
			this.rank = rank;
			this.visible = visible;
		}
		
		Card copy() {
			Card c = new Card(suit, rank, visible);
			
			return c;
		}
		
		void draw(int[] p, MPatience m, boolean mov) {
			if (mov == ((dx != 0) || (dy != 0))) {
				if (visible) backGC.drawImage(cards[suit][rank], p[0]+dx, p[1]+dy, m);
				else backGC.drawImage(back, p[0]+dx, p[1]+dy, m);
			}
		}
		
		void move() {
			double td = Math.sqrt(dx*dx + dy*dy);
			
			if (td < 50) {
				dx = 0;
				dy = 0;
			} else {
				td = td/(td - 50);
				
				dx = (int)(dx / td);
				dy = (int)(dy / td);
			}
		}
	}
	
	
	
	class Stack {
		int index;
		int x;
		int y;
		int kind;
		int num = 0;
		Card bottom;
		Card top;
		
		Stack(int i, int k) {
			index = i;
			kind = k;
			
			switch (kind) {
				case T_DECK:
					x = SCREEN_W - 108;
					y = SCREEN_H - 150;
				break;
				case T_WORKING:
					x = 10 + XOFF2 * i;
					y = 10;
				break;
				case T_DONE:
					x = 10;
					y = SCREEN_H - 150;
				break;
				case T_MOVING:
					x = 0;
					y = 0;
				break;
			}
		}
		
		Stack copy() {
			Stack s = new Stack(index, kind);
			
			Card c = bottom;
			
			while (c != null) {
				s.add(c.copy());
				c = c.next;
			}
			
			return s;
		}
		
		int[] pos(int c) {
			switch (kind) {
				case T_MOVING:	return new int[] {x, y + c * 35};
				case T_WORKING:	return new int[] {x, y + c * yoff()};
				case T_DECK:	return new int[] {x - (int)Math.floor(c/8) * XOFF1, y};
				case T_DONE:
					if (mode != 2) return new int[] {x + (int)Math.floor(c/8) * XOFF1, y};
					
					double a = timecheck - time;
					
					if (m_i == 0) {
						if (a > 1.25) return new int[] {x + (int)Math.floor(c/8) * XOFF1, y};
						if (a > 0.25) return new int[] {x + (int)Math.round(Math.floor(c/8) * XOFF1 * (a - 0.25)), y};
						if (a > 0) return new int[] {x, y};
						m_i = 1;
						f_dx = -1;
						f_dy = 1;
					}
					
					int mp = Math.max(f_px, f_py);
					
					if (a <= 0) {
						f_px = (int)Math.floor(Math.random() * 8) + 1;
						f_py = (int)Math.floor(Math.random() * 8) + 1;
						
						while (((f_px == f_py) && (f_px > 1))
							|| ((f_px % 2 == 0) && (f_py % 2 == 0))
							|| ((f_px % 3 == 0) && (f_py % 3 == 0))) f_py = (int)Math.floor(Math.random() * 8) + 1;

						mp = Math.max(f_px, f_py);

						f_sx = f_dx;
						f_sy = f_dy;
						f_dx = f_sx * ((f_px % 2 == 0) ? 1 : -1);
						f_dy = f_sy * ((f_py % 2 == 0) ? 1 : -1);
						
						m_j = (int)Math.floor(Math.random() * (mp + 2));

						timecheck = time + 1 + mp + 4 + m_j*3.5;
					}
					
					a = (timecheck - time) - 0.5 - c/16.0 - (int)(c/8) * m_j / 2.0;

					if (a < 0) return new int[] {CARD_CX + CARD_AX * f_dx, CARD_CY + CARD_AY * f_dy};
					if (a > mp) return new int[] {CARD_CX + CARD_AX * f_sx, CARD_CY + CARD_AY * f_sy};
					
					a *= Math.PI / mp;
					
					return new int[] {(int)Math.round(CARD_CX + CARD_AX * f_dx * Math.cos(a * f_px)),
									  (int)Math.round(CARD_CY + CARD_AY * f_dy * Math.cos(a * f_py))};
			}
			
			return new int[] {0, 0};
		}
		
		int yoff() {
			if (num == 0) return 35;
			return (int)Math.min(35, (SCREEN_H - 275) / num);
		}
		
		void add(Card n) {
			if (bottom == null) {
				bottom = n;
				top = n;
			} else {
				top.next = n;
				n.prev = top;
				top = n;
			}
			
			num++;
		}
		
		void deal(boolean visible, Stack dest) {
			transfer(num-1, dest);
			
			dest.top.visible = visible;
		}
		
		void transfer(int n, Stack dest) {
			if ((n < num) && (bottom != null)) {
				Card cur = bottom;
				
				for (int i=0; i<n; i++) cur = cur.next;
				
				top = cur.prev;
				
				if (top != null) top.next = null;
				else bottom = null;
				
				if (dest.bottom == null) {
					dest.bottom = cur;
				} else {
					dest.top.next = cur;
				}

				cur.prev = dest.top;
				
				for (int i=n; i<num; i++) {
					dest.top = cur;
					
					int[] tp = pos(i);
					int[] dp = dest.pos(dest.num + i-n);
					
					cur.dx = tp[0] - dp[0];
					cur.dy = tp[1] - dp[1];
					
					dest.num++;
					cur = cur.next;
				}
				
				num = n;
			}
		}
		
		void check() {
			Card c = top;
			
			if ((c != null) && (c.rank == 0)) {
				while ((c.prev != null) && c.prev.visible && (c.prev.suit == c.suit) && (c.prev.rank == c.rank + 1)) c = c.prev;
				
				if (c.rank == 7) {
					transfer(num - 8, stacks[S_DONE]);
					
					if (top != null) top.visible = true;
					
					if (stacks[S_DONE].num == 64) {
						mode = 2;
						timecheck = time + 1.5;
						m_i = 0;
					}
				}
			}
		}
		
		void draw(MPatience m, boolean mov) {
			if ((kind == T_WORKING) && !mov) backGC.drawImage(stack, x, y, m);
			
			if (bottom != null) {
				Card cur = bottom;
				int i = 0;
				
				while (cur != null) {
					cur.draw(pos(i), m, mov);
					cur = cur.next;
					i++;
				}
			}
		}
		
		void move() {
			if (bottom != null) {
				Card cur = bottom;
				
				while (cur != null) {
					cur.move();
					cur = cur.next;
				}
			}
		}
	}
	
	
	
	class State {
		Stack[] stacks = new Stack[S_MOVING + 1];
		State prev;
		
		State(Stack[] cur) {
			for (int i=0; i<=S_MOVING; i++) stacks[i] = cur[i].copy();
		}
		
		void desc() {
			String s = "";
			
			for (int i=0; i<8; i++) s += "; " + (i + 1) + ":" + stacks[S_WORKING + i].num;
			
			System.out.println("      D:" + stacks[S_DECK].num + "; C:" + stacks[S_DONE].num + s);
		}
	}
}
